@@ -1,7 +1,7 @@
 /* ------------------------------------------------------------------------- */
 
 /*
- * Copyright 2007-2017 GRAHAM DUMPLETON
+ * Copyright 2007-2020 GRAHAM DUMPLETON
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -694,12 +694,32 @@ void wsgi_log_python_error(request_rec *r, PyObject *log,
             PyObject *object = NULL;
 
             if (wsgi_event_subscribers()) {
+                WSGIThreadInfo *thread_info;
+
+                thread_info = wsgi_thread_info(0, 0);
+
                 event = PyDict_New();
 
+#if AP_MODULE_MAGIC_AT_LEAST(20100923,2)
+                if (r->log_id) {
+#if PY_MAJOR_VERSION >= 3
+                    object = PyUnicode_DecodeLatin1(r->log_id,
+                                                    strlen(r->log_id), NULL);
+#else
+                    object = PyString_FromString(r->log_id);
+#endif
+                    PyDict_SetItemString(event, "request_id", object);
+                    Py_DECREF(object);
+                }
+#endif
+
                 object = Py_BuildValue("(OOO)", type, value, traceback);
                 PyDict_SetItemString(event, "exception_info", object);
                 Py_DECREF(object);
 
+                PyDict_SetItemString(event, "request_data",
+                                     thread_info->request_data);
+
                 wsgi_publish_event("request_exception", event);
 
                 Py_DECREF(event);
